/**
 * <copyright>
 *
 * Copyright (c) 2005 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 *   IBM - Initial API and implementation
 *
 * </copyright>
 *
 * $Id$
 */
package org.eclipse.jet.compiler;


import java.util.ArrayList;
import java.util.Collections;
import java.util.List;


/**
 * Container class core JET2 AST elements contained in another AST elements in their body.
 *
 */
class BodyElements
{

  private List bodyElements = new ArrayList();

  private final JET2ASTElement owner;

  /**
   * Create a new instance 
   * @param owner The owner of the body elements
   */
  public BodyElements(JET2ASTElement owner)
  {
    super();
    this.owner = owner;
  }

  /**
   * Return a list of elements in the body
   * @return a {@link List} of elements, the empty list if there are no elements.
   */
  public List getBodyElements()
  {
    return Collections.unmodifiableList(bodyElements);
  }

  /**
   * Add a text element to the body
   * @param text the text to add
   */
  public void addBodyElement(TextElement text)
  {
    bodyElements.add(text);
    text.setParent(owner);
  }

  /**
   * Add an XML element to the body
   * @param xmlElement the xml element
   */
  public void addBodyElement(XMLElement xmlElement)
  {
    bodyElements.add(xmlElement);
    xmlElement.setParent(owner);
  }

  /**
   * Add a Java element to the body
   * @param javaElement the Java element
   */
  public void addBodyElement(JavaElement javaElement)
  {
    bodyElements.add(javaElement);
    javaElement.setParent(owner);
  }

  /**
   * Add a JET2 Directive to the body
   * @param directive the directive
   */
  public void addBodyElement(JET2Directive directive)
  {
    bodyElements.add(directive);
    directive.setParent(owner);
  }

  /**
   * Add a comment to the body.
   * @param comment the comment.
   */
  public void addBodyElement(Comment comment)
  {
    bodyElements.add(comment);
    comment.setParent(owner);
  }

  /**
   * @return Returns the owner.
   */
  public final JET2ASTElement getOwner()
  {
    return owner;
  }

  /**
   * Return the element after the passed element
   * @param element an element in the list
   * @return the next element or <code>null</code>
   * @throws IllegalArgumentException if element is not in the body
   */
  public JET2ASTElement elementAfter(JET2ASTElement element)
  {
    JET2ASTElement nextElement = null;
    if (bodyElements == null)
    {
      throw new IllegalArgumentException();
    }
    int index = bodyElements.indexOf(element);
    if (index == -1)
    {
      throw new IllegalArgumentException();
    }

    if (index + 1 < bodyElements.size())
    {
      nextElement = (JET2ASTElement)bodyElements.get(index + 1);
    }
    return nextElement;
  }

  public void addBodyElement(XMLBodyElementEnd endTag)
  {
    bodyElements.add(endTag);
    endTag.setParent(owner);
  }

  /**
   * 
   * @param element
   * @return
   */
  public JET2ASTElement elementBefore(JET2ASTElement element)
  {
    JET2ASTElement prevElement = null;
    if (bodyElements == null)
    {
      throw new IllegalArgumentException();
    }
    int index = bodyElements.indexOf(element);
    if (index == -1)
    {
      throw new IllegalArgumentException();
    }

    if (index > 0)
    {
      prevElement = (JET2ASTElement)bodyElements.get(index - 1);
    }
    return prevElement;
  }

}
