/**
 * <copyright>
 *
 * Copyright (c) 2005 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 *   IBM - Initial API and implementation
 *
 * </copyright>
 *
 * $Id$
 */

package org.eclipse.jet.compiler;


import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.emf.codegen.jet.JETException;
import org.eclipse.jet.internal.parser.JETReader;
import org.eclipse.jet.internal.parser.ParseProblemSeverity;
import org.eclipse.jet.taglib.TagLibraryReference;


/**
 * Represent a compilation unit (a template) in the JET2 AST.
 *
 */
public final class JET2CompilationUnit extends JET2ASTElement
{

  private BodyElements bodyElements = null;

  private final List problems = new ArrayList();

  private final String fileName;

  private String outputJavaPackage = null; //"org.eclipse.jet.compiled";

  private String outputJavaClassName;

  private boolean errors = false;

  private boolean warnings = false;

  private final IFile file;

  private Map predefinedLibraryMap = Collections.EMPTY_MAP;

  private TagLibraryReference[] tagLibraryReferences;

  private Set imports = new LinkedHashSet();

   public JET2CompilationUnit()
  {
    super(new JET2AST(), -1, -1, -1, -1);
    this.fileName = ""; //$NON-NLS-1$
    this.file = null;
  }

  public JET2CompilationUnit(IFile file)
  {
    super(new JET2AST(), -1, -1, -1, -1);
    this.file = file;
    this.fileName = file.getFullPath().toString();
  }

  /**
   * Define tag library prefixes (and associated tag library ids) that are
   * automatically available to the transform.
   * @param predefinedLibraryMap a map from prefix to tag library id.
   */
  public void setPredefinedTagLibraries(Map predefinedLibraryMap)
  {
    this.predefinedLibraryMap = predefinedLibraryMap;
  }
  
  public void parse(String document)
  {
    final String ENCODING = "UTF-8"; //$NON-NLS-1$
    try
    {
      InputStream is = new ByteArrayInputStream(document.getBytes(ENCODING));
      internalParse(is, ENCODING);
    }
    catch (UnsupportedEncodingException e)
    {
      // Should not happen
      throw new RuntimeException("Should not have happened", e); //$NON-NLS-1$
    }

  }

  public void parse() throws CoreException
  {
    if (file == null)
    {
      throw new IllegalStateException();
    }

    final InputStream contents = file.getContents();
    final String charset = file.getCharset();
    internalParse(contents, charset);

  }

  public void parse(InputStream contents, String charset)
  {
    internalParse(contents, charset);
  }

  /**
   * @param contents
   * @param charset
   */
  private void internalParse(final InputStream contents, final String charset)
  {
    // reset parse state...
    problems.clear();
    errors = false;
    warnings = false;
    bodyElements = null;
    outputJavaPackage = null; // "org.eclipse.jet.compiled";
    outputJavaClassName = null;

    try
    {
      JETReader reader = new JETReader(fileName, contents, charset);
      
      ASTCompilerParseListener listener = new ASTCompilerParseListener(this, reader, 
        predefinedLibraryMap != null ? predefinedLibraryMap : Collections.EMPTY_MAP);
      
      listener.parse(reader);

      tagLibraryReferences = listener.getTagLibraryReferences();
      return;
    }
    catch (JETException e)
    {
      // create a minimal compilation unit with the exeception recorded as the error.
      this.createProblem(ParseProblemSeverity.ERROR, Problem.JETException, e.getLocalizedMessage(), null, 0, 0, 1, 1);
    }
    finally
    {
      try
      {
        contents.close();
      }
      catch (IOException e)
      {
        // ignore the io exception, can't do anything about it;
      }
    }
  }

  /**
   * Return the internal BodyElements object that gives access to an updateable
   * list of AST elements the the compilation unit contains
   * @return a BodyElements object
   */
  BodyElements getInternalBodyElements()
  {
    if (bodyElements == null)
    {
      bodyElements = new BodyElements(this);
    }
    return bodyElements;
  }

  /**
   * Return a {@link List} of JET2 AST element (@link JET2ASTElement} instances.
   * @return a List. The empty list of there are no body elements.
   */
  public final List getBodyElements()
  {
    if (bodyElements == null)
    {
      return Collections.EMPTY_LIST;
    }
    else
    {
      return bodyElements.getBodyElements();
    }
  }

  /**
   * @see org.eclipse.jet.compiler.JET2ASTElement#accept(org.eclipse.jet.compiler.JET2ASTVisitor)
   */
  public void accept(JET2ASTVisitor visitor)
  {
    visitor.visit(this);
    for (Iterator i = getBodyElements().iterator(); i.hasNext();)
    {
      JET2ASTElement element = (JET2ASTElement)i.next();
      element.accept(visitor);
    }
    visitor.endVisit(this);
  }

  /**
   * Return a list of problems discovered in the compilation unit
   * @return a List of {@link Problem} objects. The empty list is returned if no problems were found.
   */
  public List getProblems()
  {
    return Collections.unmodifiableList(problems);
  }

  /**
   * Test if the compilation unit has any errors
   * @return <code>true</code> if the compilation unit had errors, <code>false</code> otherwise.
   */
  public boolean hasErrors()
  {
    return errors;
  }

  /**
   * Test if the compilation unit has any warnings
   * @return <code>true</code> if the compilation unit had warnings, <code>false</code> otherwise.
   */
  public boolean hasWarnings()
  {
    return warnings;
  }

  /**
   * Create a new problem on the compilation unit
   * @param error the severity of the problem
   * @param problemId the problem id. A value from {@link Problem} static files
   * @param message an error message, with optional replacement tokens
   * @param messageArgs the error message arguments
   * @param start the start offset of the problem (doc relative)
   * @param end the end offset of the problem (doc relative)
   * @param line the line number of the problem (1 based)
   * @param colOffset TODO
   */
  public void createProblem(ParseProblemSeverity error, int problemId, String message, Object[] messageArgs, int start, int end, int line, int colOffset)
  {
    if (error == ParseProblemSeverity.ERROR)
    {
      errors = true;
    }
    else if (error == ParseProblemSeverity.WARNING)
    {
      warnings = true;
    }
    problems.add(new Problem(fileName, error, problemId, message, messageArgs, start, end, line, colOffset));
  }

  /**
   * Return the name of the Java package to which the compilation unit will be compiled.
   * @return a string
   */
  public String getOutputJavaPackage()
  {
    return outputJavaPackage;
  }

  /**
   * Return the unqualified name of the Java class into which the compilation unit will be compiled.
   * @return Returns the outputJavaClassName.
   */
  public String getOutputJavaClassName()
  {
    return outputJavaClassName;
  }

  /**
   * Set the unqualifeid name of the Java class into which the compilation unit will be compiled.
   * @param outputJavaClassName The outputJavaClassName to set.
   */
  public void setOutputJavaClassName(String outputJavaClassName)
  {
    this.outputJavaClassName = outputJavaClassName;
  }

  /**
   * Set the Java package into which the compilation unit will be compiled.
   * @param outputJavaPackage The outputJavaPackage to set.
   */
  public void setOutputJavaPackage(String outputJavaPackage)
  {
    this.outputJavaPackage = outputJavaPackage;
  }

  public boolean removeLineWhenOtherwiseEmpty()
  {
    return false;
  }
  
  /**
   * Return an array of tag libraries referenced by this template.
   * @return a possibly empty array of tag library references.
   */
  public TagLibraryReference[] getTagLibraryReferences()
  {
    return tagLibraryReferences;
  }

  public void addImports(List list)
  {
    imports .addAll(list);
  }

  public Set getImports()
  {
    return Collections.unmodifiableSet(imports);
  }
}
