/**
 * <copyright>
 *
 * Copyright (c) 2005 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 *   IBM - Initial API and implementation
 *
 * </copyright>
 *
 * $Id$
 */

package org.eclipse.jet.compiler;


import java.io.ByteArrayInputStream;
import java.io.UnsupportedEncodingException;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.SubProgressMonitor;
import org.eclipse.jet.JET2Platform;
import org.eclipse.jet.internal.InternalJET2Platform;
import org.eclipse.jet.internal.JETActivatorWrapper;
import org.eclipse.jet.internal.compiler.JETCompilerState;
import org.eclipse.jet.internal.compiler.LoaderGenerationParameters;
import org.eclipse.jet.internal.compiler.TemplateRunner;
import org.eclipse.jet.internal.compiler.TextTrimmingVisitor;
import org.eclipse.jet.internal.l10n.JET2Messages;
import org.eclipse.jet.internal.parser.ParseProblemSeverity;
import org.eclipse.jet.taglib.TagLibraryReference;
import org.eclipse.jet.transform.IJETBundleDescriptor;
import org.eclipse.jet.transform.IJETBundleManager;


/**
 * The compiler that turns a JET2 sources files Java files in the Eclipse workspace.
 *
 */
public class JET2Compiler
{

  public static final String RUNTIME_PROBLEM_MARKER = JET2Platform.PLUGIN_ID + ".runtimeProblem"; //$NON-NLS-1$

  /**
   * The Marker created by the compiler on all identified problems.
   */
  public static final String COMPILE_PROBLEM_MARKER = JET2Platform.PLUGIN_ID + ".compileProblem"; //$NON-NLS-1$

  /**
   * The default value of the  {@link #OPTION_COMPILED_TEMPLATE_PACKAGE}  option; value: "org.eclipse.jet2.internal.compiled".
   */
  public static final String DEFAULT_COMPILED_TEMPLATE_PACKAGE = "org.eclipse.jet.compiled"; //$NON-NLS-1$

  /**
   * The default value of the  {@link #OPTION_COMPILED_TEMPLATE_SRC_DIR}  option; value: "jet2java".
   */
  public static final String DEFAULT_COMPILED_TEMPLATE_SRC_DIR = "jet2java"; //$NON-NLS-1$

  /**
   * The default value of the  {@link #OPTION_TEMPLATE_EXT}  option; value: "jet".
   */
  public static final String DEFAULT_TEMPLATE_EXT = "jet"; //$NON-NLS-1$
  
  public static final Boolean DEFAULT_SET_JAVA_FILES_AS_DERIVED = Boolean.TRUE;

  private static final String NS = "org.eclipse.jet."; //$NON-NLS-1$

  /**
   * Compiler option specifying the package to which compiled templates are written; value: "compiledTemplatePackage".
   */
  public static final String OPTION_COMPILED_TEMPLATE_PACKAGE = NS + "compiledTemplatePackage"; //$NON-NLS-1$

  /**
   * Compiler option specifying the Java project source directory to which compiled templates are written; value: "compiledTemplateSrcDir".
   */
  public static final String OPTION_COMPILED_TEMPLATE_SRC_DIR = NS + "compiledTemplateSrcDir"; //$NON-NLS-1$

  /**
   * Compiler option specifying which extensions are recognized as extensions; value: "templateExt".
   */
  public static final String OPTION_TEMPLATE_EXT = NS + "templateExt"; //$NON-NLS-1$
  
  public static final String OPTION_SET_JAVA_FILES_AS_DERIVED = NS + "setJavaDerived"; //$NON-NLS-1$

  private static Map defaultCompileOptions = null;
  private static final Object defaultCompileOptionsLock = new Object();
  
  // Return the default compile options...
  public static Map getDefaultCompilerOptions()
  {
    if(defaultCompileOptions == null)
    {
      Map map = new HashMap();
      map.put(OPTION_COMPILED_TEMPLATE_PACKAGE, DEFAULT_COMPILED_TEMPLATE_PACKAGE);
      map.put(OPTION_COMPILED_TEMPLATE_SRC_DIR, DEFAULT_COMPILED_TEMPLATE_SRC_DIR);
      map.put(OPTION_TEMPLATE_EXT, DEFAULT_TEMPLATE_EXT);
      map.put(OPTION_SET_JAVA_FILES_AS_DERIVED, DEFAULT_SET_JAVA_FILES_AS_DERIVED);
      
      synchronized(defaultCompileOptionsLock)
      {
        defaultCompileOptions = Collections.unmodifiableMap(map);
      }
    }
    return defaultCompileOptions;
  }
  
  private final IProgressMonitor monitor;

  private final IProject project;

  private IJETCompilerState compilerState;

  private int noCompiles;

  private final Map options;

  private final Map predefinedLibraryMap;

  private final String templateLoaderPackage;

  private final String templateLoaderClassName;
  
  private Set sourceExtensions = null;

  private final TemplateRunner templateRunner;

  /**
   * Create an instance of the JET compiler capable of incremental compilation from a saved state.
   * @param project the project containing templates to be compiled.
   * @param options the compiler options
   * @param compilerState the saved state from previous a previous compile
   * @param monitor a progress monitor
   */
  public JET2Compiler(IProject project, Map options, IJETCompilerState compilerState, IProgressMonitor monitor)
  {
    super();
    this.project = project;
    this.monitor = monitor == null ? new NullProgressMonitor() : monitor;
    this.options = options == null ? getDefaultCompilerOptions() : options;
    this.compilerState = compilerState;

    this.noCompiles = 0;

    final IJETBundleManager bundleManager = JET2Platform.getJETBundleManager();

    final IJETBundleDescriptor descriptor = bundleManager.getDescriptorForProject(project.getName());
    predefinedLibraryMap = descriptor != null ? getPredefinedPrefixMap(descriptor) : Collections.EMPTY_MAP;

    if (descriptor != null && descriptor.getTemplateLoaderClass() != null)
    {
      final String templateLoaderClass = descriptor.getTemplateLoaderClass();
      int index = templateLoaderClass.lastIndexOf('.');
      templateLoaderPackage = index >= 0 ? templateLoaderClass.substring(0, index) : ""; //$NON-NLS-1$
      templateLoaderClassName = index >= 0 ? templateLoaderClass.substring(index + 1) : templateLoaderClass;
    }
    else
    {
      templateLoaderPackage = templateLoaderClassName = null;
    }
    templateRunner = new TemplateRunner();
  }
  /**
   * Create an instance
   * @param project the project upon which the compiler will act
   * @param monitor a progress monitor for compiler actions 
   * @param options a map of compiler options
   * 
   */
  public JET2Compiler(IProject project, IProgressMonitor monitor, Map options)
  {
    this(project, options, new JETCompilerState(), monitor);
  }

  private Map getPredefinedPrefixMap(IJETBundleDescriptor descriptor)
  {
    final TagLibraryReference[] tlRefs = descriptor.getTagLibraryReferences();
    final Map result = new HashMap(tlRefs.length);
    for (int i = 0; i < tlRefs.length; i++)
    {
      if (tlRefs[i].isAutoImport())
      {
        result.put(tlRefs[i].getPrefix(), tlRefs[i].getTagLibraryId());
      }
    }
    return result;
  }

  /**
   * Create the Template Loader class for the project 
   */
  public void createTransformFiles()
  {
    if(noCompiles > 0)
    {
      createTemplateLoaderClass();
    }
  }

  /**
   * Create the Class that loads compiled templates.
   */
  private void createTemplateLoaderClass()
  {
    if(templateLoaderClassName == null) return;
    
    final LoaderGenerationParameters args = new LoaderGenerationParameters(templateLoaderPackage, 
      templateLoaderClassName, compilerState.getTemplateMap());
    
    String src = templateRunner.generate("templates/jet2transform.jet",  //$NON-NLS-1$
      Collections.singletonMap("args", args)); //$NON-NLS-1$

    IPath targetPath = new Path(getSourceDirOption()).append(
      templateLoaderPackage.replace('.', '/')).append(templateLoaderClassName).addFileExtension("java"); //$NON-NLS-1$

    try
    {
      final IFile file = project.getFile(targetPath);
      writeFile(src, file, file.getCharset());
    }
    catch (CoreException e)
    {
      InternalJET2Platform.logError(JET2Messages.JET2Compiler_ErrorWritingJava, e);
    }
    catch (UnsupportedEncodingException e)
    {
      InternalJET2Platform.logError(JET2Messages.JET2Compiler_ErrorWritingJava, e);
    }
  }

  /**
   * @return
   */
  private String getSourceDirOption()
  {
    return getOption(JET2Compiler.OPTION_COMPILED_TEMPLATE_SRC_DIR, JET2Compiler.DEFAULT_COMPILED_TEMPLATE_SRC_DIR);
  }

  private String getOption(String key, String defaultValue)
  {
    Object value = options.get(key);
    return value == null ? defaultValue : value.toString(); 
  }
  
  private boolean getBooleanOption(String key, boolean defaultValue)
  {
    Object value = options.get(key);
    return value == null ? defaultValue : 
      value instanceof Boolean ? ((Boolean)value).booleanValue() :
        Boolean.valueOf(value.toString()).booleanValue();
  }
  
  private boolean isJETSource(IPath path)
  {
    if(sourceExtensions == null) {
      sourceExtensions = new HashSet();
      final String[] extensions = getOption(OPTION_TEMPLATE_EXT, DEFAULT_TEMPLATE_EXT).split(","); //$NON-NLS-1$
      for (int i = 0; i < extensions.length; i++)
      {
        sourceExtensions.add(extensions[i]);
      }
    }
    return sourceExtensions.contains(path.getFileExtension());
  }
  /**
   * Test whether the supplied resource is a compilable JET2 source file.
   * @param resource
   * @return <code>true</code> if the identified resource is a jet2 source file
   */
  public boolean isJET2SourceFile(IResource resource)
  {
    return resource.getType() == IResource.FILE && isJETSource(resource.getFullPath());
  }

  /**
   * Compile the passed resource into a Java file. Does nothing if the passed resource
   * is not a JET2 template.
   * @param resource a project resource
   */
  public void compile(IResource resource)
  {
    if (!isJET2SourceFile(resource))
    {
      return; // ignore
    }

    ++noCompiles;

    IFile file = (IFile)resource;

    clean(resource.getProjectRelativePath().toString());
    
    removeMarkers(resource);

    JET2CompilationUnit cu = new JET2CompilationUnit(file);
    cu.setPredefinedTagLibraries(predefinedLibraryMap);
    try
    {
      cu.parse();

      cu.accept(new TagValidationVisitor(cu));

      markProblems(resource, cu);

      if (!cu.hasErrors())
      {
        cu.accept(new TextTrimmingVisitor());
        writeCompiledOutput(file, cu);
      }
    }
    catch (CoreException e)
    {
      // core exceptions represent problems reading the source file
      createParseExceptionMarker(resource, e);
    }

  }
  /**
   * @param resource
   */
  private void removeMarkers(IResource resource)
  {
    try
    {
      resource.deleteMarkers(COMPILE_PROBLEM_MARKER, false, IResource.DEPTH_INFINITE);
      resource.deleteMarkers(RUNTIME_PROBLEM_MARKER, false, IResource.DEPTH_INFINITE);
    }
    catch (CoreException e)
    {
      InternalJET2Platform.logError(JET2Messages.JET2Compiler_ErrorRemovingMarkers, e);
    }
  }

  private void createParseExceptionMarker(IResource resource, CoreException e)
  {
    try
    {
      IMarker marker = resource.createMarker(COMPILE_PROBLEM_MARKER);
      marker.setAttribute(IMarker.MESSAGE, e.getStatus().getMessage());
      marker.setAttribute(IMarker.PRIORITY, IMarker.PRIORITY_HIGH);
      marker.setAttribute(IMarker.SEVERITY, IMarker.SEVERITY_ERROR);
    }
    catch (CoreException ex)
    {
      InternalJET2Platform.logError(JET2Messages.JET2Compiler_ErrorCreatingMarkers, e);
    }
  }

  /**
   * Place problem markers on the passed resource for each problem recorded in the compilation unit.
   * @param resource the resource to mark
   * @param cu the compilation unit corresponding to the resource
   */
  private void markProblems(IResource resource, JET2CompilationUnit cu)
  {
    try
    {
      for (Iterator i = cu.getProblems().iterator(); i.hasNext();)
      {
        Problem problem = (Problem)i.next();
        IMarker marker = resource.createMarker(COMPILE_PROBLEM_MARKER);
        marker.setAttribute(IMarker.LINE_NUMBER, problem.getLineNumber());
        marker.setAttribute(IMarker.CHAR_START, problem.getStart());
        marker.setAttribute(IMarker.CHAR_END, problem.getEnd());
        marker.setAttribute(IMarker.MESSAGE, problem.getMessage());
        marker.setAttribute(IMarker.PRIORITY, IMarker.PRIORITY_HIGH);
        if (problem.getSeverity() == ParseProblemSeverity.ERROR)
        {
          marker.setAttribute(IMarker.PRIORITY, IMarker.PRIORITY_HIGH);
          marker.setAttribute(IMarker.SEVERITY, IMarker.SEVERITY_ERROR);
        }
        else if (problem.getSeverity() == ParseProblemSeverity.WARNING)
        {
          marker.setAttribute(IMarker.PRIORITY, IMarker.PRIORITY_NORMAL);
          marker.setAttribute(IMarker.SEVERITY, IMarker.SEVERITY_WARNING);
        }
        else
        {
          marker.setAttribute(IMarker.PRIORITY, IMarker.PRIORITY_LOW);
          marker.setAttribute(IMarker.SEVERITY, IMarker.SEVERITY_INFO);
        }
      }
    }
    catch (CoreException e)
    {
      InternalJET2Platform.logError(JET2Messages.JET2Compiler_ErrorWritingJava, e);
    }
  }

  /**
   * @param file
   * @param cu
   */
  private void writeCompiledOutput(IFile file, JET2CompilationUnit cu)
  {

    ensureJavaOutputSet(file, cu);

    final String templatePath = file.getProjectRelativePath().toString();
    final String qualifiedClassName = makeFullyQualifiedJavaName(cu.getOutputJavaPackage(), cu.getOutputJavaClassName());

    if (templatePath.equals(compilerState.getTemplateFromClass(qualifiedClassName)))
    {
      cu.createProblem(
        ParseProblemSeverity.ERROR,
        Problem.MultipleTemplatesWithSameJavaClass,
        JET2Messages.JET2Compiler_SameJavaClassAsOther,
        new Object []{ qualifiedClassName, compilerState.getTemplateFromClass(qualifiedClassName), },
        -1,
        -1,
        -1,
        -1);
    }

    String javaSrc = templateRunner.generate("templates/jet2java.jet",  //$NON-NLS-1$
      Collections.singletonMap("cu", cu)); //$NON-NLS-1$

    IFile targetFile = file.getProject().getFile(javaClassProjectPath(qualifiedClassName));
    try
    {
      writeFile(javaSrc, targetFile, file.getCharset());
      compilerState.addTemplate(templatePath, qualifiedClassName);
      compilerState.addTemplateTagLibraryReferences(templatePath, cu.getTagLibraryReferences());
    }
    catch (CoreException e)
    {
      InternalJET2Platform.logError(JET2Messages.JET2Compiler_ErrorWritingJava, e);
    }
    catch (UnsupportedEncodingException e)
    {
      InternalJET2Platform.logError(JET2Messages.JET2Compiler_ErrorWritingJava, e);
    }
  }

  /**
   * Ensure the the compilation unit has Java output package and class 
   * set. If not, generate names for these, taking into account that
   * a name may have been generated for them already.
   * @param file
   * @param cu
   */
  private boolean ensureJavaOutputSet(IFile file, JET2CompilationUnit cu)
  {
    boolean inferedJavaPackage = false;
    final String templatePath = file.getProjectRelativePath().toString();

    // See if we already have a name for this...
    final String existingQualifiedClass = (String)compilerState.getTemplateMap().get(templatePath);

    if (existingQualifiedClass == null)
    {
      // we haven't generated a name before, do it now...
      
      if (cu.getOutputJavaPackage() == null)
      {
        cu.setOutputJavaPackage(getOption(JET2Compiler.OPTION_COMPILED_TEMPLATE_PACKAGE, JET2Compiler.DEFAULT_COMPILED_TEMPLATE_PACKAGE));
        inferedJavaPackage = true;
      }
      if (cu.getOutputJavaClassName() == null)
      {
        // check for conflicts because we have name mangled
        // or for identically named files in different directories.
        String baseClassName = makeJavaClassName(file.getFullPath().removeFileExtension().lastSegment());

        String className = baseClassName;
        String potentialQualifiedName = makeFullyQualifiedJavaName(cu.getOutputJavaPackage(), className);

        for (int i = 0; compilerState.getTemplateFromClass(potentialQualifiedName) != null; i++)
        {
          className = baseClassName + "_" + i; //$NON-NLS-1$
          potentialQualifiedName = makeFullyQualifiedJavaName(cu.getOutputJavaPackage(), className);
        }

        cu.setOutputJavaClassName(className);
      }
    }
    else
    {
      int index = existingQualifiedClass.lastIndexOf('.');
      String existingPackage = index == -1 ? "" : existingQualifiedClass.substring(0, index); //$NON-NLS-1$
      String existingClass = index == -1 ? existingQualifiedClass : existingQualifiedClass.substring(index + 1);

      if (cu.getOutputJavaPackage() == null)
      {
        cu.setOutputJavaPackage(existingPackage);
      }
      if (cu.getOutputJavaClassName() == null)
      {
        cu.setOutputJavaClassName(existingClass);
      }
    }
    return inferedJavaPackage;
  }

  /**
   * @param javaClassName
   * @return the project relative path of the generated java file
   */
  private IPath javaClassProjectPath(String javaClassName)
  {
    IPath targetPath = new Path(getSourceDirOption()).append(javaClassName.replace('.', '/')).addFileExtension("java"); //$NON-NLS-1$
    return targetPath;
  }

  private String makeFullyQualifiedJavaName(String javaPackage, String javaClassName)
  {
    return javaPackage == null || javaPackage.trim().length() == 0 ? javaClassName : javaPackage + '.' + javaClassName;
  }

  /**
   * @param javaSrc
   * @param targetFile
   * @param charset the character encoding of the .jet source file
   * @throws CoreException
   * @throws UnsupportedEncodingException 
   */
  private void writeFile(String javaSrc, IFile targetFile, String charset) throws CoreException, UnsupportedEncodingException
  {
    ensureDirsExist(targetFile.getParent());

    if (!targetFile.exists())
    {
      targetFile.create(new ByteArrayInputStream(new byte []{}), true, new SubProgressMonitor(monitor, 1));
    }

    if (!targetFile.getCharset().equals(charset))
    {
      targetFile.setCharset(charset, new SubProgressMonitor(monitor, 1));
    }

    targetFile.setContents(new ByteArrayInputStream(javaSrc.getBytes(charset)), true, false, new SubProgressMonitor(monitor, 1));
    targetFile.setDerived(getBooleanOption(OPTION_SET_JAVA_FILES_AS_DERIVED, DEFAULT_SET_JAVA_FILES_AS_DERIVED.booleanValue()));
  }

  private void ensureDirsExist(IContainer container) throws CoreException
  {
    if (container != null && !container.exists() && container.getType() == IResource.FOLDER)
    {
      ensureDirsExist(container.getParent());
      ((IFolder)container).create(false, true, monitor);
    }
  }

  private String makeJavaClassName(String name)
  {
    StringBuffer result = new StringBuffer("_jet_"); //$NON-NLS-1$
    for (int i = 0; i < name.length(); i++)
    {
      char c = name.charAt(i);
      if (Character.isJavaIdentifierPart(c))
      {
        result.append(c);
      }
    }
    return result.toString();
  }

  /**
   * Remove any resources for the passed resource representing a JET template.
   * The method will remove any Java files created by the compiler for the resource.
   * <p>
   * This method is equivalent to:
   * <code>removeDerivedResources(resource, false)</code>.
   * 
   * @param resource a resource
   * @throws CoreException if a derived resource cannot be removed.
   * @see #removeDerivedResources(IResource, boolean)
   */
  public void removeDerivedResources(IResource resource) throws CoreException
  {
    removeDerivedResources(resource, false);
  }
  /**
   * Remove any derived resources for this resource
   * @param resource a resource
   * @param derivedOnly if <code>true</code>, Java files with the derived attribute set are removed.
   * @throws CoreException if a derived resource cannot be removed.
   */
  public void removeDerivedResources(IResource resource, boolean derivedOnly) throws CoreException
  {
    if (!isJET2SourceFile(resource))
    {
      return; // ignore
    }

    String javaClassName = (String)compilerState.getTemplateMap().get(resource.getProjectRelativePath().toString());
    if (javaClassName != null)
    {
      ++noCompiles;
      removeJavaClass(javaClassName, derivedOnly);
    }

  }
  /**
   * @param javaClassName
   * @param derivedOnly
   * @throws CoreException
   */
  private void removeJavaClass(String javaClassName, boolean derivedOnly) throws CoreException
  {
    IPath javaSrcFile = javaClassProjectPath(javaClassName);

    IFile file = project.getFile(javaSrcFile);
    if (file.exists() && (file.isDerived() || !derivedOnly))
    {
      file.delete(true, new SubProgressMonitor(monitor, 1));
    }
  }

  /**
   * @return Returns the noCompiles.
   */
  public final int getNoCompiles()
  {
    return noCompiles;
  }
  
  public void clean() 
  {
    // make a safe copy of the keyset, as we will be modifying the template map by cleaning...
    final Set templatePathSet = new HashSet(compilerState.getTemplateMap().keySet());
    monitor.beginTask("", templatePathSet.size() + 2); //$NON-NLS-1$
    for (Iterator i = templatePathSet.iterator(); i.hasNext();)
    {
      String templatePath = (String)i.next();
      clean(templatePath);
    }
    
    if(templateLoaderClassName != null)
    {
      try
      {
        removeJavaClass(templateLoaderPackage + "." + templateLoaderClassName, true); //$NON-NLS-1$
      }
      catch (CoreException e)
      {
        JETActivatorWrapper.INSTANCE.log(e);
      }
    }
    
    removeMarkers(project);
    monitor.done();
  }
  /**
   * @param templatePath
   */
  private void clean(String templatePath)
  {
    IFile templateResource = project.getFile(templatePath);
    try
    {
      removeDerivedResources(templateResource, true);
      compilerState.removeTemplate(templatePath);
    }
    catch (CoreException e)
    {
      JETActivatorWrapper.INSTANCE.log(e);
    }
  }
  
  /**
   * Returns the current compiler state
   * @return the current compiler state
   */
  public final IJETCompilerState getCompilerState()
  {
    return compilerState;
  }
  
  
}
