/**
 * <copyright>
 *
 * Copyright (c) 2006 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 *   IBM - Initial API and implementation
 *
 * </copyright>
 *
 * $Id$
 */
package org.eclipse.jet.internal;

import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.ISaveContext;
import org.eclipse.core.resources.ISaveParticipant;
import org.eclipse.core.resources.ISavedState;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Platform;

/**
 * Manager for state saved and restored by the jet plugin
 */
public class SavedStateManager implements ISaveParticipant
{
  private static boolean DEBUG = InternalJET2Platform.getDefault().isDebugging()
  && Boolean.valueOf(Platform.getDebugOption("org.eclipse.jet/debug/savedState")).booleanValue(); //$NON-NLS-1$

  public static interface IStateSaver 
  {

    /**
     * The name of the file to which state information will be saved as
     * a relative path.
     * @return a relative path
     */
    String getStateFileName();

    /**
     * Save the state to the location (absolute path) provided
     * @param location the location (absolute path) of the saved state file
     */
    void doSave(IPath location) throws CoreException;

    /**
     * The project for which the saver is saving data
     * @return
     */
    IProject getProject();
  }
  
  private ISavedState savedState;
  private List stateSavers = new ArrayList();
  private Map saversByProject = new HashMap();
  private final InternalJET2Platform plugin;

  /**
   * 
   */
  public SavedStateManager()
  {
    super();
    plugin = InternalJET2Platform.getDefault();
  }

  /* (non-Javadoc)
   * @see org.eclipse.core.resources.ISaveParticipant#doneSaving(org.eclipse.core.resources.ISaveContext)
   */
  public void doneSaving(ISaveContext context)
  {
  }

  /* (non-Javadoc)
   * @see org.eclipse.core.resources.ISaveParticipant#prepareToSave(org.eclipse.core.resources.ISaveContext)
   */
  public void prepareToSave(ISaveContext context) throws CoreException
  {
  }

  /* (non-Javadoc)
   * @see org.eclipse.core.resources.ISaveParticipant#rollback(org.eclipse.core.resources.ISaveContext)
   */
  public void rollback(ISaveContext context)
  {
    
  }

  /* (non-Javadoc)
   * @see org.eclipse.core.resources.ISaveParticipant#saving(org.eclipse.core.resources.ISaveContext)
   */
  public void saving(ISaveContext context) throws CoreException
  {
    switch (context.getKind())
    {
      case ISaveContext.FULL_SAVE:
        // save the plug-in state
        doSave(context, stateSavers);
        break;
      case ISaveContext.PROJECT_SAVE:
        // get the project related to this save operation
        IProject project = context.getProject();
        List projectSavers = (List)saversByProject.get(project);
        if(projectSavers != null) {
          doSave(context, projectSavers);
        }
        break;
      case ISaveContext.SNAPSHOT:
        // This operation needs to be really fast because
        // snapshots can be requested frequently by the
        // workspace.
        // TODO Is this fast enough?
        doSave(context, stateSavers);
        break;
    }
  }

  /**
   * @param context
   * @param projectSavers
   * @throws IllegalStateException
   * @throws CoreException
   */
  private void doSave(ISaveContext context, List projectSavers) throws IllegalStateException, CoreException
  {
    for (Iterator i = projectSavers.iterator(); i.hasNext();)
    {
      final WeakReference ref = (WeakReference)i.next();
      final IStateSaver saver = (IStateSaver)ref.get();
      if(saver != null) 
      {
        IPath logicalPath = getLogicalStatePath(saver);
      
        final IPath realPath = getActualPath(logicalPath, context.getSaveNumber());
        context.map(logicalPath, realPath);
        
        final IPath location = plugin.getStateLocation().append(realPath);
  
        saver.doSave(location);
        context.needSaveNumber();
      }
    }
  }

  /**
   * @param path
   * @param saveNumber
   * @return
   */
  private IPath getActualPath(IPath logicalPath, int saveNumber)
  {
    IPath saveDirPath = getSaveDirectory(saveNumber);
    
    final IPath realPath = saveDirPath.append(logicalPath);
    return realPath;
  }

  /**
   * @param saver
   * @return
   */
  private IPath getLogicalStatePath(IStateSaver saver)
  {
    IPath logicalPath = new Path(saver.getProject().getName()).append(saver.getStateFileName());
    return logicalPath;
  }

  /**
   * Return the directory to which the state will be saved
   * @param saveNumber the save number
   * @return a path relative to the plug-in's workspace save location
   */
  private Path getSaveDirectory(int saveNumber)
  {
    return new Path("save-" + Integer.toString(saveNumber)); //$NON-NLS-1$
  }

  public void startup()
  {
    if(DEBUG) System.out.println("SavedStateManager.startup()"); //$NON-NLS-1$
    try
    {
      savedState = ResourcesPlugin.getWorkspace().addSaveParticipant(InternalJET2Platform.getDefault(), this);
      if(DEBUG && savedState != null) System.out.println("   found saved state: " + savedState.getSaveNumber()); //$NON-NLS-1$
    }
    catch (CoreException e)
    {
      savedState = null;
    }
  }
  
  public void shutdown()
  {
    if(DEBUG) System.out.println("SavedStateManager.shutdown()"); //$NON-NLS-1$
    ResourcesPlugin.getWorkspace().removeSaveParticipant(InternalJET2Platform.getDefault());
  }
  
  public IPath addSaveSaver(IStateSaver saver ,IProject project)
  {
    stateSavers.add(new WeakReference(saver));
    rememberProjectSaver(project, saver);
    IPath initialLocation = null;
    if(savedState != null)
    {
      final int saveNumber = savedState.getSaveNumber();
      final IPath realPath = getActualPath(getLogicalStatePath(saver), saveNumber);
      initialLocation = InternalJET2Platform.getDefault().getStateLocation().append(realPath);
    }
    if(DEBUG) System.out.println("SavedStateManager.addSaveSaver(): " + project + ", " + initialLocation);  //$NON-NLS-1$//$NON-NLS-2$
    return initialLocation;
  }
  
  private void rememberProjectSaver(IProject project, IStateSaver saver)
  {
    // FIXME This method appears to do nothing. Is it even needed?
    List list = (List)saversByProject.get(project);
    if(list == null)
    {
      list = new ArrayList();
    }
    list.add(new WeakReference(saver));
  }

}
