/**
 * <copyright>
 *
 * Copyright (c) 2005 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 *   IBM - Initial API and implementation
 *
 * </copyright>
 *
 * $Id$
 */

package org.eclipse.jet.internal.builder;


import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.jar.JarFile;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.resources.IResourceDeltaVisitor;
import org.eclipse.core.resources.IResourceVisitor;
import org.eclipse.core.resources.IncrementalProjectBuilder;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.SubProgressMonitor;
import org.eclipse.jet.JET2Platform;
import org.eclipse.jet.compiler.CompileOptionsManager;
import org.eclipse.jet.compiler.IJETCompilerState;
import org.eclipse.jet.compiler.JET2Compiler;
import org.eclipse.jet.internal.InternalJET2Platform;
import org.eclipse.jet.internal.JETActivatorWrapper;
import org.eclipse.jet.internal.SavedStateManager;
import org.eclipse.jet.internal.compiler.JETCompilerState;
import org.eclipse.jet.internal.l10n.JET2Messages;
import org.eclipse.osgi.util.NLS;


/**
 * Implement the builder that compiles JET2 files into Java
 *
 */
public class JET2Builder extends IncrementalProjectBuilder implements SavedStateManager.IStateSaver
{
  private static boolean DEBUG = InternalJET2Platform.getDefault().isDebugging()
  && Boolean.valueOf(Platform.getDebugOption("org.eclipse.jet/debug/builder")).booleanValue(); //$NON-NLS-1$

  static final String COMPILESTATE_PROJECT_DIR = "projects"; //$NON-NLS-1$

  private int filesToCompileCount;
  
  private JETCompilerState compilerState;
  
  private final SavedStateManager savedStateManager;

  /**
   * 
   */
  public JET2Builder()
  {
    super();
    savedStateManager = InternalJET2Platform.getDefault().getSavedStateManager();
  }

  protected void clean(IProgressMonitor monitor) throws CoreException
  {
    if(DEBUG) System.out.println("JET2Builder: cleaning " + getProject());     //$NON-NLS-1$
    final JET2Compiler compiler = getJETCompiler(new SubProgressMonitor(monitor, 1));
    
    compiler.clean();
    
    compilerState.clear();
    if(DEBUG) System.out.println("*JET2Builder: done cleaning " + getProject());     //$NON-NLS-1$
  }
  
  /* (non-Javadoc)
   * @see org.eclipse.core.resources.IncrementalProjectBuilder#build(int, java.util.Map, org.eclipse.core.runtime.IProgressMonitor)
   */
  protected IProject[] build(int kind, Map args, IProgressMonitor monitor) throws CoreException
  {
    if (DEBUG)
      System.out.println("JET2Builder: build(" + getProject() + ", " + kind + ")"); //$NON-NLS-1$//$NON-NLS-2$//$NON-NLS-3$
    try
    {
      if (kind == IncrementalProjectBuilder.CLEAN_BUILD)
      {
        compilerState.clear();
      }
      else if (kind == IncrementalProjectBuilder.FULL_BUILD)
      {
        clean(new SubProgressMonitor(monitor, 1));
        fullBuild(new SubProgressMonitor(monitor, 1));
      }
      else
      {
        
        IResourceDelta delta = getDelta(getProject());
        if (isReferencedProjectChanged() || delta == null)
        {
          clean(new SubProgressMonitor(monitor, 1));
          fullBuild(new SubProgressMonitor(monitor, 1));
        }
        else
        {
          incrementalBuild(delta, monitor);
        }
      }
      return computeReferencedProjects(compilerState);
    }
    finally
    {
      if (DEBUG)
        System.out.println("JET2Builder: done build(" + getProject() + ", " + kind + ")"); //$NON-NLS-1$//$NON-NLS-2$//$NON-NLS-3$
    }
  }

  private boolean isReferencedProjectChanged()
  {
    final IProject[] projects = computeReferencedProjects(compilerState);
    for (int i = 0; i < projects.length; i++)
    {
      final IResourceDelta delta = getDelta(projects[i]);
      if(delta != null)
      {
        if(delta.findMember(new Path("plugin.xml")) != null  //$NON-NLS-1$
            || delta.findMember(new Path(JarFile.MANIFEST_NAME)) != null)
        {
          if(DEBUG) System.out.println("   forcing full build. ref updated:" + projects[i]); //$NON-NLS-1$
          return true;
        }
      }
    }
    
    return false;
  }

  /**
   * Perform an incremental build
   * @param mainDelta
   * @param monitor
   * @throws CoreException
   */
  private void incrementalBuild(IResourceDelta mainDelta, final IProgressMonitor monitor) throws CoreException
  {
    if(DEBUG) System.out.println("JET2Builder: incrementalBuild " + getProject());     //$NON-NLS-1$
    compilerState.setProjectTagLibraryReferences(JET2Platform.getProjectDescription(getProject().getName()).getTagLibraryReferences());
    final JET2Compiler compiler = getJETCompiler(monitor);
    filesToCompileCount = 0;
    // figure out how much work we must do (for the progress monitor).
    mainDelta.accept(new IResourceDeltaVisitor()
      {
        public boolean visit(IResourceDelta delta) throws CoreException
        {
          switch (delta.getKind())
          {
            case IResourceDelta.ADDED:
            case IResourceDelta.CHANGED:
            case IResourceDelta.REMOVED:
              if (compiler.isJET2SourceFile(delta.getResource()))
              {
                filesToCompileCount++;
              }
              break;
          }
          return true;
        }
      });
    monitor.beginTask(NLS.bind(JET2Messages.JET2Builder_CompilingChanged, getProject().getName()), filesToCompileCount);
    try
    {
      mainDelta.accept(new IResourceDeltaVisitor()
        {

          public boolean visit(IResourceDelta delta) throws CoreException
          {
            final IResource resource = delta.getResource();
            switch (delta.getKind())
            {
              case IResourceDelta.ADDED:
              case IResourceDelta.CHANGED:
                if(DEBUG) System.out.println("JET2Builder: add/change " + resource);                   //$NON-NLS-1$
                if(resource.getType() == IResource.FILE && getProject().equals(resource.getProject())) {
                  monitor.subTask(NLS.bind(JET2Messages.JET2Builder_Compiling, resource.getProjectRelativePath()));
                  compiler.compile(resource);
                }
                monitor.worked(1);
                break;
              case IResourceDelta.REMOVED:
                if(DEBUG) System.out.println("JET2Builder remove " + resource);                   //$NON-NLS-1$
                monitor.subTask(NLS.bind(JET2Messages.JET2Builder_Cleaning, resource.getProjectRelativePath()));
                if(resource.getType() == IResource.FILE && getProject().equals(resource.getProject())) {
                  compiler.removeDerivedResources(resource, false);
                  compilerState.removeTemplate(resource.getProjectRelativePath().toString());
                }
                monitor.worked(1);
                break;
            }
            return true;
          }
        });
      compiler.createTransformFiles();

    }
    catch (CoreException e)
    {
      compilerState.clear();
      forgetLastBuiltState();
      throw e;
    }
    finally
    {
      monitor.done();
      if(DEBUG) System.out.println("JET2Builder: done incrementalBuild " + getProject());     //$NON-NLS-1$
    }

  }

  /**
   * Perform a Full build
   * @param monitor
   * @throws CoreException
   */
  private void fullBuild(final IProgressMonitor monitor) throws CoreException
  {
    if(DEBUG) System.out.println("JET2Builder: fullBuild " + getProject());     //$NON-NLS-1$
    compilerState.setProjectTagLibraryReferences(JET2Platform.getProjectDescription(getProject().getName()).getTagLibraryReferences());
    final JET2Compiler compiler = getJETCompiler(monitor);
    filesToCompileCount = 0;
    // figure out how much work we must do (for the progress monitor).
    getProject().accept(new IResourceVisitor()
      {
        public boolean visit(IResource resource) throws CoreException
        {
          if (compiler.isJET2SourceFile(resource))
          {
            filesToCompileCount++;
          }
          return true;
        }
      });
    monitor.beginTask(NLS.bind(JET2Messages.JET2Builder_CompilingAll, getProject().getName()), filesToCompileCount);
    try
    {
      getProject().accept(new IResourceVisitor()
        {

          public boolean visit(IResource resource) throws CoreException
          {
            monitor.subTask(NLS.bind(JET2Messages.JET2Builder_Compiling, resource.getProjectRelativePath()));
            compiler.compile(resource);
            monitor.worked(1);
            return true;
          }
        });
      compiler.createTransformFiles();
      // TODO Save compilation state upon successful build
    }
    catch (CoreException e)
    {
      compilerState.clear();
      forgetLastBuiltState();
      throw e;
    }
    finally
    {
      monitor.done();
      if(DEBUG) System.out.println("JET2Builder: done fullBuild " + getProject());     //$NON-NLS-1$
    }
  }

  /**
   * @param monitor
   * @return
   */
  private JET2Compiler getJETCompiler(final IProgressMonitor monitor)
  {
    final JET2Compiler compiler = new JET2Compiler(getProject(), 
            CompileOptionsManager.getOptions(getProject()),
            compilerState,
            monitor);
    return compiler;
  }

  /*
   * Load saved compiler state...
   */
  protected void startupOnInitialize()
  {
    super.startupOnInitialize();

    compilerState = null;

    if (DEBUG) System.out.println("JET2Builder: startupOnInitialize " + getProject()); //$NON-NLS-1$
    try
    {
      IPath fileLocation = savedStateManager.addSaveSaver(this, getProject());
      
      if (fileLocation != null)
      {
        File savedStateFile = fileLocation.toFile();
        try
        {
          ObjectInputStream ois = new ObjectInputStream(new java.io.FileInputStream(savedStateFile));
          compilerState = (JETCompilerState)ois.readObject();
          ois.close();
          if (DEBUG)
          {
            System.out.println("    loaded compilerState from saved state"); //$NON-NLS-1$
            System.out.print("         taglibs   : ["); //$NON-NLS-1$
            String[] ids = compilerState.getAllReferencedTagLibraryIds();
            for (int i = 0; i < ids.length; i++)
            {
              if (i != 0)
                System.out.print(", "); //$NON-NLS-1$
              System.out.print(ids[i]);
            }
            System.out.println("]"); //$NON-NLS-1$
            System.out.println("         templates : " + compilerState.getTemplateMap()); //$NON-NLS-1$
          }
        }
        catch (FileNotFoundException e)
        {
          // nothing to do, the state was not saved for some reason.
          // NO NOT log this, this will happen whenever a new project occurs.
        }
        catch (IOException e)
        {
          // error reading the file, assume state is unrecoverable.
          // remove the state file.
          JETActivatorWrapper.INSTANCE.log(e);
          savedStateFile.delete();
        }
        catch (ClassNotFoundException e)
        {
          // error in deserializing the file. This should not happen. Log an error, for the record...
          InternalJET2Platform.logError("Could not deserialize JET2Builder State", e); //$NON-NLS-1$
        }

      }
    }
    finally
    {
      if (compilerState == null)
      {
        if(DEBUG) System.out.println("    compilerState not loaded from saved state"); //$NON-NLS-1$
        // don't have the saved build state, force a full rebuild...
        compilerState = new JETCompilerState();
        forgetLastBuiltState();
      }
      if (DEBUG)
        System.out.println("JET2Builder: done startupOnInitialize " + getProject()); //$NON-NLS-1$
    }
  }
  
  private IProject[] computeReferencedProjects(IJETCompilerState compilerState)
  {
    final String[] allReferencedTagLibraryIds = compilerState.getAllReferencedTagLibraryIds();
    List referencedProjects = new ArrayList(allReferencedTagLibraryIds.length);
    for (int i = 0; i < allReferencedTagLibraryIds.length; i++)
    {
      final IProject refProject = InternalJET2Platform.getDefault().getProjectDefiningTagLibrary(allReferencedTagLibraryIds[i]);
      if(refProject != null)
      {
        referencedProjects.add(refProject);
        if(DEBUG) System.out.println("     depends on:" + refProject); //$NON-NLS-1$
      }
    }
    return (IProject[])referencedProjects.toArray(new IProject[referencedProjects.size()]);
  }

  public IJETCompilerState getCompilerState()
  {
    return compilerState;
  }
  
  public void doSave(IPath location) throws CoreException
  {
    FileOutputStream fileStream = null;
    ObjectOutputStream objStream = null;
    try
    {
      
      File file = location.toFile();
      file.getParentFile().mkdirs();
      fileStream = new FileOutputStream(file);
      objStream = new ObjectOutputStream(fileStream);
      objStream.writeObject(compilerState);
    }
    catch (IOException e)
    {
      throw new CoreException(InternalJET2Platform.newStatus(IStatus.ERROR, e.getLocalizedMessage(), e));
    }
    finally
    {
      if(objStream != null)
      {
        try
        {
          objStream.close();
        }
        catch (IOException e)
        {
          JETActivatorWrapper.INSTANCE.log(e);
        }
      }
      if(fileStream != null)
      {
        try
        {
          fileStream.close();
        }
        catch (IOException e)
        {
          JETActivatorWrapper.INSTANCE.log(e);
        }
      }
    }
  }

  public String getStateFileName()
  {
    return "compilerState.tmp"; //$NON-NLS-1$
  }

}
