/**
 * <copyright>
 *
 * Copyright (c) 2006 IBM Corporation and others.
 * All rights reserved.   This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: 
 *   IBM - Initial API and implementation
 *
 * </copyright>
 *
 * $Id$
 */
package org.eclipse.jet.taglib;

import java.text.MessageFormat;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.jet.IWriterListener;
import org.eclipse.jet.JET2Writer;
import org.eclipse.jet.internal.l10n.JET2Messages;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.IRegion;
import org.eclipse.jface.text.Position;

/**
 * Helper for implementing task markers on writer output.
 */
public class MarkerHelper implements IWriterListener
{

  private static final String TASKMARKER_ID = "org.eclipse.jet.taskmarker"; //$NON-NLS-1$

  private static final class MarkerPosition extends Position
  {

    public final String description;
    public final TagInfo tagInfo;
    public final String templatePath;

    public MarkerPosition(int start, int length, String description, TagInfo tagInfo, String templatePath)
    {
      super(start, length);
      this.description = description;
      this.tagInfo = tagInfo;
      this.templatePath = templatePath;
    }
    
  }
  
  private static final String MARKER_CATEGORY = MarkerHelper.class.getName();
  /**
   * 
   */
  private MarkerHelper()
  {
    // do nothing
  }

  public void finalizeContent(JET2Writer writer, Object file) throws JET2TagException
  {
    // nothing to do...
  }

  public void postCommitContent(JET2Writer writer, Object file) throws JET2TagException
  {
    if(file instanceof IFile)
    {
      IFile iFile = (IFile)file;
      try
      {
        iFile.deleteMarkers(TASKMARKER_ID, false, IResource.DEPTH_ZERO);
        Position[] positions = writer.getPositions(MARKER_CATEGORY);
        for (int i = 0; i < positions.length; i++)
        {
          MarkerPosition markerPos = (MarkerPosition)positions[i];
          
          final IMarker marker = iFile.createMarker(TASKMARKER_ID);
          int line = writer.getDocument().getLineOfOffset(markerPos.offset);
          int start = markerPos.offset;
          int end = markerPos.offset + markerPos.length;

          // for empty marked regions, turn this into a marker on the line.
          if(start == end)
          {
            final IRegion lineInformation = writer.getDocument().getLineInformation(line);
            start = lineInformation.getOffset();
            end = start + lineInformation.getLength();
          }
          
          String msg = markerPos.description;
          if(msg == null || msg.length() == 0)
          {
            msg = writer.getDocument().get(start, end - start);
          }
          marker.setAttribute(IMarker.LINE_NUMBER, line);
          marker.setAttribute(IMarker.CHAR_START, start);
          marker.setAttribute(IMarker.CHAR_END, end);
          marker.setAttribute(IMarker.MESSAGE, msg);
//          marker.setAttribute(IMarker.PRIORITY, IMarker.PRIORITY_HIGH);
        }
      }
      catch (CoreException e)
      {
        final String msg = JET2Messages.MarkerHelper_CouldNotCreateMarker;
        throw new JET2TagException(MessageFormat.format(msg, new Object[] {iFile.getFullPath().toString()}), e);
      }
      catch (BadLocationException e)
      {
        throw new JET2TagException(e);
      }
      
    }
  }

  public static void createMarkerOnWriter(JET2Writer writer, int start, int end, String description, TagInfo tagInfo, String templatePath)
  {
    writer.addPositionCategory(MARKER_CATEGORY);
    final MarkerPosition markerPosition = new MarkerPosition(start, end - start, description, tagInfo, templatePath);
    writer.addPosition(MARKER_CATEGORY, markerPosition);
    writer.addEventListener(MARKER_CATEGORY, new MarkerHelper());
  }
}
